#include <Windows.h>
#include <Shlwapi.h>
#include <Psapi.h>
#include <process.h>
#include <map>
#include <string>

#include "resource.h"

#pragma comment(lib, "Shlwapi.lib")
#pragma comment(lib, "psapi.lib")


DWORD GetOpenName(TCHAR * outbuf, const TCHAR * filter, const TCHAR * title);

HINSTANCE ghInstance = NULL;
HWND ghWnd;
std::map<std::wstring, HMODULE> loaded_dlls;

INT_PTR CALLBACK MainDlgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    unsigned int pos;
    HMODULE dll, mods[1024];
    TCHAR path[MAX_PATH], dll_name[32];

    switch(uMsg)
    {
        case WM_INITDIALOG:
            ghWnd = hWnd;

            for(std::map<std::wstring, HMODULE>::const_iterator it = loaded_dlls.begin(); it != loaded_dlls.end(); ++it)
                SendDlgItemMessage(hWnd, IDC_LIST, LB_ADDSTRING, 0, reinterpret_cast<LPARAM>(it->first.c_str()));

            SwitchToThisWindow(hWnd, FALSE);
        break;

        case WM_COMMAND:
            switch(LOWORD(wParam))
            {
                case IDC_LOAD:
                    if(GetOpenName(path, L"DLL (*.dll)\0*.dll\0All files (*.*)\0*.*\0\0", L"Select dll to load..."))
                    {
                        if( (dll = GetModuleHandle(path)) != NULL)
                        {
                            MessageBox(hWnd, L"Can't load same DLL twice", L"Error", MB_OK | MB_ICONERROR);
                            break;
                        }

                        if( (dll = LoadLibrary(path)) == NULL)
                        {
                            MessageBox(hWnd, L"Can't load specified DLL", L"Error", MB_OK | MB_ICONERROR);
                            break;
                        }

                        GetModuleBaseName(GetCurrentProcess(), dll, dll_name, sizeof(dll_name) / sizeof(dll_name[0]));
                        
                        SendDlgItemMessage(hWnd, IDC_LIST, LB_ADDSTRING, 0, reinterpret_cast<LPARAM>(dll_name));
                        
                        loaded_dlls.insert(std::make_pair(dll_name, dll));
                    }
                break;

                case IDC_UNLOAD:
                    pos = SendDlgItemMessage(hWnd, IDC_LIST, LB_GETCURSEL, NULL, NULL);
                    if(pos != LB_ERR && loaded_dlls.size() > 0)
                    {
                        SendDlgItemMessage(hWnd, IDC_LIST, LB_GETTEXT, static_cast<WPARAM>(pos), reinterpret_cast<LPARAM>(dll_name));

                        if(FreeLibrary(loaded_dlls[dll_name]) == NULL)
                        {
                            MessageBox(hWnd, L"Can't unload specified DLL", L"Error", MB_OK | MB_ICONERROR);
                            break;
                        }

                        loaded_dlls.erase(dll_name);
                        SendDlgItemMessage(hWnd, IDC_LIST, LB_DELETESTRING, static_cast<WPARAM>(pos), 0);
                    }
                break;

                case IDC_REFRESH:
                    DWORD cb;
                    HANDLE pr = GetCurrentProcess();

                    if(EnumProcessModules(pr, mods, sizeof(mods), &cb))
                    {
                        std::map<std::wstring, HMODULE>::const_iterator it = loaded_dlls.begin();

                        while(it != loaded_dlls.end())
                        {
                            bool is_present = false;

                            for(unsigned int i = 0; i < cb / sizeof(HMODULE); i++)
                                if(GetModuleBaseName(pr, mods[i], dll_name, sizeof(dll_name) / sizeof(dll_name[0])))
                                    if(it->first == dll_name)
                                        is_present = true;


                            if(!is_present)
                            {
                                pos = SendDlgItemMessage(hWnd, IDC_LIST, LB_FINDSTRINGEXACT, static_cast<WPARAM>(-1), reinterpret_cast<LPARAM>(it->first.c_str()));

                                if(pos != LB_ERR)
                                    SendDlgItemMessage(hWnd, IDC_LIST, LB_DELETESTRING, static_cast<WPARAM>(pos), 0);

                                loaded_dlls.erase(it++);
                            }
                            else
                                ++it;
                        }
                    }
                    else
                    {
                        MessageBox(hWnd, L"Can't get process modules list", L"Error", MB_OK | MB_ICONERROR);
                    }
                break;
            }
        break;

        case WM_CLOSE:
            EndDialog(hWnd, 0);
        break;

        default:
            return 0;
    }

    return 1;
}

unsigned int __stdcall thread_proc(void * arg)
{
    if(!RegisterHotKey(NULL, 1, MOD_ALT | MOD_NOREPEAT, 'M'))
    {
        MessageBox(HWND_DESKTOP, L"Failed to register hotkey", L"Error", MB_OK | MB_ICONERROR);
        return 1;
    }

    MSG msg = {0};

    while(GetMessage(&msg, NULL, 0, 0) != 0)
    {
        if(msg.message == WM_HOTKEY)
        {
            DialogBoxParam(ghInstance, MAKEINTRESOURCE(IDD_MAIN), 0, MainDlgProc, 0);
        }
    }
    
    return 0;
}

BOOL WINAPI DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID reserved)
{
    static HANDLE thread = NULL;

    if(dwReason == DLL_PROCESS_ATTACH)
    {
        if(!ghInstance)
            ghInstance = hInstance;

        if(!thread)
            thread = reinterpret_cast<HANDLE>(_beginthreadex(NULL, 0, &thread_proc, NULL, 0, NULL));

    }
    else if(dwReason == DLL_PROCESS_DETACH)
    {
        CloseHandle(thread);
    }
 
    return TRUE;
}

DWORD GetOpenName(TCHAR * outbuf, const TCHAR * filter, const TCHAR * title)
{
    OPENFILENAME ofn = {0};
    TCHAR buf[MAX_PATH + 2];
 
    GetModuleFileName(NULL, buf, MAX_PATH);
 
    TCHAR * tmp = StrRChr(buf, NULL, L'\\');
    if(tmp != 0)
    {
        *tmp = 0;
        ofn.lpstrInitialDir = buf;
    }
 
    ofn.hInstance = ghInstance;
    ofn.hwndOwner = ghWnd;
    ofn.lStructSize = sizeof(OPENFILENAME);
    ofn.lpstrFilter = filter;
    ofn.nFilterIndex = 1;
    ofn.lpstrFile = outbuf;
    ofn.lpstrFile[0] = 0;
    ofn.lpstrFile[1] = 0;
    ofn.nMaxFile = MAX_PATH;
    ofn.lpstrTitle = title;
    ofn.Flags = OFN_EXPLORER | OFN_DONTADDTORECENT | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_LONGNAMES | OFN_NONETWORKBUTTON | OFN_PATHMUSTEXIST;
 
    return GetOpenFileName(&ofn);
}

extern "C" __declspec(dllexport) void __cdecl dummy(void)
{
}
